unit SuggEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls, DBCtrls, Graphics,
  Forms, StdCtrls, ComCtrls;

type
  TACListBox = class(TCustomListBox)
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  end;

  PTDictionary = ^TDictionary;
  TDictionary = record
    Name: string;              // Names this entry. Optional.
    Path: string;              // Where to get/write the data entries
    RefCount: DWORD;           // Number of controls referring to me
    bDirty: boolean;           // There are new words not saved
    bSave: boolean;            // Save on exit
    ListaValores: TStringList; // Actual list of words of this dictionary
  end;

  TDictionaryList = class(TObject)
    Lista: TList;              // List of dictionaries hold

  protected
    function FindPath(dicPath: string): integer;

  public
    constructor Create;
    destructor  Destroy; override;

    function Add(dicName, dicPath: string; SaveOnExit: Boolean = True): integer;
    procedure Delete(dicName: string); overload;
    procedure Delete(ID: integer; save: boolean = True); overload;
    function Find(dicName: string): integer;
    function GetDict(ID: integer): TStringList; overload;
    function GetDict(Name: string): TStringList; overload;

    function AddWord(ID: integer; word: string): boolean; overload;
    function AddWord(Name: string; word: string): boolean; overload;
  end;

  TShowBoxEvent = procedure(Sender: TObject; count: DWord) of object;

  TAutoCompleteEdit = class(TCustomEdit)
  private
    DictID: integer;            // ID del diccionario
    bSuggest: Boolean;          // Deseamos sugerencias
    FFichero: string;           // Fichero de diccionario
    iPos: integer;              // Sugerencia seleccionada
    iGrupo: integer;            // Cabecera del grupo que nos interesa
    uLines: word;               // Lineas de la sugerencia
    DefButton: TButton;         // Botn por defecto del formulario propietario
    backColor: TColor;          // Background Color

    bIgnoreChange: bool;        // Evitar recursin
    bFromList: bool;            // El campo se ha actualizado de la lista

 // Selects if the Suggestions' Listbox is owned by the control's parent or by
 // the holding form. The difference is visual: if owned by the control's
 // parent AND the control is on a panel then the listbox will be clipped
 // by at the panel's limits. If owned by the form it will span the panel limits
 // and overlay every other component of the form (but a status bar by the way)
    bSpanParent: bool;          // Exceder el control padre ?

    FOnShowBox: TShowBoxEvent;  // Al mostrar la ventana de sugerencias
    FOnHideBox: TNotifyEvent;   // Al ocultar la ventana

  protected
    procedure SetSuggest(val: boolean);
    procedure SetSuggestHeight(val: word);
    procedure SetDictionary(path: string);
    procedure CalcSuggestPos(ControlRect: TRect);
    procedure SetColor(val: TColor);
    function  StatusBarHeight(Form: TCustomForm): integer;
    procedure SetSpan(val: bool);

    // Funciones de apoyo que pueden ser sustituidas
    function  GetGroupIndex(c: char): integer; virtual;
    procedure OverrideDefault; virtual;
    procedure ReleaseDefault; virtual;

    // Procedimientos originales modificados
    procedure CNKeyDown(var Message: TWMKEYDOWN); message WM_KEYDOWN;
    procedure CNKeyUp(var Message: TWMKEYUP); message WM_KEYUP;
    procedure WMWindowPosChanged(var Message: TWMWINDOWPOSCHANGED); message WM_WINDOWPOSCHANGED;

    procedure Change; override;
    procedure DoExit; override;
    procedure DoEnter; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published
    // Especficos del control
    property Dictionary: string read FFichero write SetDictionary;
    property Suggest: Boolean read bSuggest write SetSuggest;
    property SuggestHeight: word read uLines write SetSuggestHeight;
    property OnShowBox: TShowBoxEvent read FOnShowBox write FOnShowBox;
    property OnHideBox: TNotifyEvent read FOnHideBox write FOnHideBox;
    property SuggestBoxColor: TColor read backColor write SetColor;
    property SpanParent: bool read bSpanParent write SetSpan;

    // Del control estandar
    property Anchors;
    property AutoSelect;
    property AutoSize;
    property BiDiMode;
    property BorderStyle;
    property CharCase;
    property Color;
    property Constraints;
    property Ctl3D;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property OEMConvert;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PasswordChar;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnChange;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDock;
    property OnStartDrag;
  end;

  TAutoCompleteDBEdit = class(TDBEdit)
  private
    DictID: integer;            // ID del diccionario
    bSuggest: Boolean;          // Deseamos ventana de sugerencias
    FFichero: string;           // Fichero de sugerencias
    iPos: integer;              // Sugerencia seleccionada
    iGrupo: integer;            // Posicin inicial de diccionario
    uLines: word;               // Lineas de la sugerencia
    DefButton: TButton;         // Botn por defecto del formulario padre
    backColor: TColor;          // Background Color

    bIgnoreChange: bool;        // Evitar recursin
    bFromList: bool;            // El campo se ha actualizado de la lista

 // Selects if the Suggestions' Listbox is owned by the control's parent or by
 // the holding form. The difference is visual: if owned by the control's
 // parent AND the control is on a panel then the listbox will be clipped
 // by at the panel's limits. If owned by the form it will span the panel limits
 // and overlay every other component of the form (but a status bar by the way)
    bSpanParent: bool;          // Exceder el control padre ?

    FOnShowBox: TShowBoxEvent;  // Al mostrar la ventana de sugerencias
    FOnHideBox: TNotifyEvent;   // Al ocultar la ventana

  private
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;

  protected
    procedure SetSuggest(val: boolean);
    procedure SetSuggestHeight(val: word);
    procedure SetDictionary(path: string);
    procedure CalcSuggestPos(ControlRect: TRect);
    procedure SetColor(val: TColor);
    function  StatusBarHeight(Form: TCustomForm): integer;
    procedure SetSpan(val: bool);

    // Funciones de apoyo que pueden ser sustituidas
    function  GetGroupIndex(c: char): integer; virtual;
    procedure OverrideDefault; virtual;
    procedure ReleaseDefault; virtual;

    // Procedimientos originales modificados
    procedure CNKeyDown(var Message: TWMKEYDOWN); message WM_KEYDOWN;
    procedure CNKeyUp(var Message: TWMKEYUP); message WM_KEYUP;
    procedure WMWindowPosChanged(var Message: TWMWINDOWPOSCHANGED); message WM_WINDOWPOSCHANGED;

    procedure Change; override;
    procedure DoExit; override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published
    // Especficos del control
    property Dictionary: string read FFichero write SetDictionary;
    property Suggest: Boolean read bSuggest write SetSuggest;
    property SuggestHeight: word read uLines write SetSuggestHeight;
    property OnShowBox: TShowBoxEvent read FOnShowBox write FOnShowBox;
    property OnHideBox: TNotifyEvent read FOnHideBox write FOnHideBox;
    property SuggestBoxColor: TColor read backColor write SetColor;
    property SpanParent: bool read bSpanParent write SetSpan;
  end;


const
  DELTA = 1; // Puntos entre la lista y el control (points between list and control)

var
  // Slo existe una nica ventana de sugerencias en todo el sistema ya que
  // slo se puede mostrar una ventana de sugerencias cada vez. Lo que hacemos
  // es mover la ventana de sugerencias al lado del control que estemos
  // editando, ajustarla a su tamao, rellenarla y hacerla visible.
  //
  // There is a system wide suggestion box because we are showing only one
  // suggestion box at a time. What we do is place the suggestion box by the
  // control we are inputting data to, adjust its size, fill it and show it.
  ListaSugerencias: TACListBox = nil;
  DictList: TDictionaryList = nil;

procedure Register;

implementation

{ TDictionaryList }

constructor TDictionaryList.Create;
begin
  inherited;

  Lista := TList.Create;
end;

destructor TDictionaryList.Destroy;
var
  i : integer;
  p: PTDictionary;
begin
    for i := 0 to Lista.Count - 1 do begin
      p := Lista[i];
      if (p^.bSave) and (p^.bDirty) and (p^.ListaValores.Count > 0) then
        p^.ListaValores.SaveToFile(p^.Path);
      p^.ListaValores.Free;
      Dispose(p);
    end;
  Lista.Free;

  inherited;
end;

function TDictionaryList.Add(dicName, dicPath: string; SaveOnExit: Boolean = True): integer;
var
  p: PTDictionary;
  dir: string;
  f: file;
  Buffer: array[0..MAX_PATH - 1] of Char;

begin
  // Add the new dictionary if it does not exist already. Identical dictionaries
  // are those that refer to the same absolute filename. Param 'dicPath' is
  // either an absolute or a relative path to a file with .\ and ..\ in it and
  // we have to assume that is correctly built but may be not.

  result :=-1; // Prepare for worse case
  GetCurrentDirectory(sizeof(Buffer), Buffer); // Save current dir.

  if StrScan(PChar(dicPath), ':') <> nil then // We have an absolute path
  begin
    dir := ExtractFilePath(dicPath);  // Get the path, no filename
    if not SetCurrentDirectory(PChar(dir)) then // The given path does exist?
    begin                                          // No, Try to create it
      if not CreateDirectory(PChar(dir), nil) then
        exit      // !!ERROR Invalid path name
      else begin  // The path could be created so it was good *AND* it had not
                  // been created before so there is no dictionay match
        SetCurrentDirectory(PChar(dir)); // Get a clean name, no dots
        dir := GetCurrentDir + '\' + ExtractFileName(dicPath);
        dicPath := dir;
        result := -1; // New dict
      end;
    end else begin                       // Yes it exists, so it is a valid one
      dir := GetCurrentDir + '\' + ExtractFileName(dicPath); // Get a clean name, no dots
      dicPath := dir;
      result := FindPath(dicPath); // Seek for a match
    end;
  end else begin // we have a relative path. Make it absolute
    dir := ExtractFileDir(dicPath);
    if not SetCurrentDirectory(PChar(dir)) then
      exit      // !!ERROR  Bad relative path. No fix for this
    else begin
      dir := GetCurrentDir + '\' + ExtractFileName(DicPath); // Get a clean name, no dots
      dicPath := dir;
      result := FindPath(dicPath); // Seek for a match
    end;
  end;

  SetCurrentDirectory(Buffer);  // Let all be the same as before

  // If we got a match, increase the reference count...
  if result <> -1 then begin
    Inc(TDictionary(Lista[result]^).RefCount);
    exit;
  end;

  // ... otherwise add it new.
  new (p);
  if (p = nil) then exit;

  if dicName = '' then dicName := 'D' + IntToStr(Lista.Count);
  p^.Name := dicName;
  p^.Path := dicPath;
  p^.ListaValores := TStringList.Create;
  p^.ListaValores.Sorted := True;
  p^.ListaValores.Duplicates := dupIgnore;
  p^.bSave := SaveOnExit;
  p^.bDirty := False;
  p^.RefCount := 0;
  Inc(p^.RefCount);
  if p^.ListaValores = nil then begin
    Dispose(p);
    exit;
  end;

  if not FileExists(dicPath) then begin
    AssignFile(f, dicPath);
    Rewrite(f);
    CloseFile(f);
  end else begin
    p^.ListaValores.LoadFromFile(dicPath);
  end;
  result := Lista.Add(p);
  if result = -1 then begin
    p^.ListaValores.Free;
    dispose(p);
  end;
end;

procedure TDictionaryList.Delete(dicName: string);
var
  i: integer;
  p: PTDictionary;
begin
  i := Find (dicName);
  if i <> -1 then begin
    p := Lista[i];
    Dec(p^.RefCount);
    if p^.RefCount = 0 then begin
      if (p^.bSave) and (p^.bDirty) and (p^.ListaValores.Count > 0) then
        p^.ListaValores.SaveToFile(p^.Path);
      Dispose(p);
      Lista.Delete(i);
      Lista.Pack;
    end;
  end;
end;

procedure TDictionaryList.Delete(ID: integer; save: boolean = True);
var
  p: PTDictionary;
begin
  if (ID >= 0) and (ID < Lista.Count) then begin
    p := Lista[ID];
    Dec(p^.RefCount);
    if p^.RefCount = 0 then begin
      if (p^.bSave) and (p^.bDirty) and (p^.ListaValores.Count > 0) then
        p^.ListaValores.SaveToFile(p^.Path);
      Dispose(p);
      Lista.Delete(ID);
      Lista.Pack;
    end;
  end;
end;

function TDictionaryList.Find(dicName: string): integer;
var
  i: integer;
begin
  result := -1;

  for i := 0 to Lista.Count - 1 do
    if AnsiCompareText(TDictionary(Lista[i]^).Name, dicName) = 0 then begin
      result := i;
      exit;
    end;
end;

function TDictionaryList.FindPath(dicPath: string): integer;
var
  i: integer;
begin
  result := -1;

  for i := 0 to Lista.Count - 1 do
    if AnsiCompareText(TDictionary(Lista[i]^).Path, dicPath) = 0 then begin
      result := i;
      exit;
    end;
end;

function TDictionaryList.GetDict(ID: integer): TStringList;
begin
  result := nil;
  if (ID >= 0) and (ID < Lista.Count) then
    result := TDictionary(Lista[ID]^).ListaValores;
end;

function TDictionaryList.GetDict(Name: string): TStringList;
var
  i: Cardinal;
begin
  result := nil;

  for i := 0 to Lista.Count - 1 do
    if TDictionary(Lista[i]^).Name = Name then begin
      result := TDictionary(Lista[i]^).ListaValores;
      exit;
    end;
end;

function TDictionaryList.AddWord(ID: integer; word: string): boolean;
var
  p: PTDictionary;
begin
  result := false;
  if (ID >= 0) and (ID < Lista.Count) then begin
    p := Lista[ID];
    result := (p^.ListaValores.Add(word) <> -1);
    p^.bDirty := p^.bDirty or result;
  end;
end;

function TDictionaryList.AddWord(Name: string; word: string): boolean;
var
  ID: integer;
begin
  result := false;
  ID := Find(Name);
  if (ID <> -1) then result := AddWord(ID, word);
end;

{ TACListBox }

// No queremos las barras de scroll normales.
// We do not want the standard scroll bars
procedure TACListBox.CreateParams(var Params: TCreateParams);
begin
  inherited;
  Params.Style := Params.Style and not (WS_VSCROLL + WS_HSCROLL);
end;


{ TAutoCompleteEdit }

constructor TAutoCompleteEdit.Create(AOwner: TComponent);
begin
  inherited;

  uLines := 3;            // Suggestion box is 3 lines height by default
  bSuggest := True;       // By default do suggest
  backColor := RGB(255,255,180); // Kind of yellow

  iPos := -1;

  if (ListaSugerencias = nil) and not (csDesigning in ComponentState) then begin
    // Dejar que otro gestione la destruccin de la lista
    // Let other handle the disposal of the Listbox
    ListaSugerencias := TACListBox.Create(AOwner);
    ListaSugerencias.Parent := Self;
    ListaSugerencias.Name := 'AutoSuggestList';
    ListaSugerencias.Visible := False;
    ListaSugerencias.MultiSelect := False;
    ListaSugerencias.IntegralHeight := True;
    ListaSugerencias.TabStop := False;
    ListaSugerencias.Height := uLines * ListaSugerencias.ItemHeight;
    ListaSugerencias.Color := backColor;
  end;

  DictID := -1;
  if not (csDesigning in ComponentState) then begin
    if DictList = nil then DictList := TDictionaryList.Create;
    if FFichero <> '' then
      DictID := DictList.Add('', FFichero, True);
  end;
end;

destructor TAutoCompleteEdit.Destroy;
begin
  if not (csDesigning in ComponentState) then // No queremos grabar los valores de diseo
    DictList.Delete(DictID);

  inherited;
end;

procedure TAutoCompleteEdit.SetColor(val: TColor);
begin
  if val <> backColor then begin
    backColor := val;
    if not (csDesigning in ComponentState) then
      ListaSugerencias.Color := val;
  end;
end;

function TAutoCompleteEdit.GetGroupIndex(c: char): integer;
var
 p1: PAnsiChar;
 L: TStringList;
begin
  L := DictList.GetDict(DictID);
  if c = '' then begin
    result := L.Count;
    exit;   // Avoid searching for nothing
  end;

  p1 := @c;
  for result := 0 to L.Count - 1 do begin
   try
     if CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
                        p1, 1, PChar(L[result]), 1) = 2 then exit;
   except
     break;
   end;
  end;
end;

procedure TAutoCompleteEdit.SetSuggestHeight(val: word);
begin
  if val <> uLines then begin
    uLines := val;
    if not (csDesigning in ComponentState) then begin
      ListaSugerencias.Height := uLines * ListaSugerencias.ItemHeight;
      if GetFocus = Handle then CalcSuggestPos(BoundsRect);
    end;
  end;
end;

procedure TAutoCompleteEdit.SetSuggest(val: boolean);
begin
  if val <> bSuggest then begin
    bSuggest := val;
    if not (csDesigning in ComponentState) then begin
      if val and not ListaSugerencias.Visible then Change;
      if not val then
        ListaSugerencias.Visible := False;
        if Assigned(FOnHideBox) then FOnHideBox(Self);
    end;
  end;
end;

procedure TAutoCompleteEdit.SetSpan(val: bool);
begin
  if val <> bSpanParent then begin
    bSpanParent := val;
  end;
end;

procedure TAutoCompleteEdit.SetDictionary(path: string);
begin
  if (FFichero <> path) then begin
    FFichero := path;
    if not (csDesigning in ComponentState) then
    begin
      DictList.Delete(DictId);
      DictID := DictList.Add('', path, True);
      if ListaSugerencias.Visible then Change;
    end;
  end;
end;

procedure TAutoCompleteEdit.WMWindowPosChanged(var Message: TWMWINDOWPOSCHANGED);
var
  rect: TRect;
begin
  inherited;

  if not (csDesigning in ComponentState) then begin
    rect.Top    := PWINDOWPOS(Message.WindowPos).y;
    rect.Bottom := rect.Top + PWINDOWPOS(Message.WindowPos).cy;
    rect.Left   := PWINDOWPOS(Message.WindowPos).x;
    rect.Right  := rect.top + PWINDOWPOS(Message.WindowPos).cx;
    CalcSuggestPos(rect);
  end;
end;

// El usuario ha pulsado una tecla especial. Nos interesan las flechas
// arriba y abajo para movernos por la lista de sugerencias y el retorno
// para seleccionar la sugerencia y hacerla activa.
//
// The user has pressed an special key. We are interested in up and down arrows
// to move through the suggestions list and return to select the actual one.
//
procedure TAutoCompleteEdit.CNKeyDown(var Message: TWMKeyDown);
begin
  if not (csDesigning in ComponentState) then begin
    if not ListaSugerencias.Visible then inherited
    else
    with Message do begin
      Result := 1;
      case CharCode of
        VK_UP, VK_DOWN, VK_RETURN, VK_ESCAPE: exit;
      end;
      Result := 0;
      inherited;
    end;
  end else inherited;
end;

procedure TAutoCompleteEdit.CNKeyUp(var Message: TWMKeyUp);
begin
  if not (csDesigning in ComponentState) then begin
    if not ListaSugerencias.Visible then inherited
    else
    with Message do begin
      Result := 0;
      case CharCode of
        VK_UP:
        begin
          if iPos = -1 then exit;
          if iPos > 0 then Dec(iPos)
                      else begin
                        iPos := -1;
                        ReleaseDefault;
                      end;
          ListaSugerencias.ItemIndex := iPos;
        end;

        VK_DOWN:
        begin
          if iPos = -1 then begin
                         iPos := 0;
                         OverrideDefault;
                       end else inc(iPos);
          if iPos >= ListaSugerencias.Items.Count then
            Dec(iPos);
          ListaSugerencias.ItemIndex := iPos;
        end;

        VK_RETURN:
        if ((iPos <> -1) and (iPos < ListaSugerencias.Items.Count)) then
        begin
          // Esto va a disparar un mensaje de cambio. Hacer que no nos moleste
          // This is firing a Change message, so don't let it disturb us...
          bIgnoreChange := True;
          Text := ListaSugerencias.Items[iPos];
          bFromList := True;
          bIgnoreChange := False;
          ListaSugerencias.Visible := False;
          if Assigned(FOnHideBox) then FOnHideBox(Self);
          SelStart := Length(Text);
          ReleaseDefault;
        end;

        VK_ESCAPE:
        begin
          ListaSugerencias.Visible := False;
          if Assigned(FOnHideBox) then FOnHideBox(Self);
          ReleaseDefault;
        end;
      end;
    end;
  end else inherited;
end;

// El contenido del control ha cambiado. Actualizar la lista de sugerencias
//
// The control's content has changed. Update list with suggestions
procedure TAutoCompleteEdit.Change;
var
  idx: integer;
  L: word;
  Lista: TStringList;

begin
  if not (csDesigning in ComponentState) and (GetFocus = Handle) and
     (DictID <> -1) and bSuggest and not bIgnoreChange then begin
    ReleaseDefault; iPos := -1;
    ListaSugerencias.Clear;
    L := Length(Text);
    if L > 0 then begin
      // Saltar al grupo con la letra inicial.
      // Jump to starting letter group
      if (L = 1) or (iGrupo = 0) then iGrupo := GetGroupIndex(Text[1]);
      Lista := DictList.GetDict(DictID);
      if Lista <> nil then
        for idx := iGrupo to Lista.Count - 1 do
        begin
          if CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
                           PChar(Lista[idx]), 1, PChar(Text), 1) = 3 then
            break; // Nos hemos pasado de letra
          if L > Length(Lista[idx]) then
            continue; // Una subcadena
          if CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
                           PChar(Text), L, PChar(Lista[idx]), L) = 2 then
            ListaSugerencias.Items.Add(Lista[idx]);
        end;

      if (ListaSugerencias.Items.Count > 0) then begin
        ListaSugerencias.Visible := True;
        ListaSugerencias.BringToFront;
        if Assigned(FOnShowBox) then FOnShowBox(Self, ListaSugerencias.Items.Count);
      end else begin
        ListaSugerencias.Visible := False;
        if Assigned(FOnHideBox) then FOnHideBox(Self);
      end;
    end else begin
      ListaSugerencias.Visible := False;
      if Assigned(FOnHideBox) then FOnHideBox(Self);
    end;
  end;

  inherited;
end;

procedure TAutoCompleteEdit.DoEnter;
begin
  if not (csDesigning in ComponentState) then begin
    // Asociarla al contexto correspondiente
    // Link to corresponding context
    if bSpanParent then ListaSugerencias.Parent := GetParentForm(Self)
                   else ListaSugerencias.Parent := Parent;

    // Ajustarla al control de que se trate
    // Adjust to control's size
    CalcSuggestPos(BoundsRect);
  end;

  inherited;
end;

procedure TAutoCompleteEdit.DoExit;
begin
  if not (csDesigning in ComponentState) then begin
    // Salvar el valor actual del control en la lista asociada
    // si no ha sido un valor obtenido de ella.
    if bSuggest and not bFromList and (Length(Text) > 0) then
      DictList.AddWord(DictID, Text);
    bFromList := False;

    if ListaSugerencias.Visible then begin
      ListaSugerencias.Visible := False;
      if Assigned(FOnHideBox) then FOnHideBox(Self);
    end;
  end;

  inherited;
end;

// Localizar el botn por defecto y cancelar su comportamiento.
//  Limitaciones: Slo cancela el primer botn por defecto localizado . Si hay
//    ms de una en un formulario (algo errneo pero parece ser que posible)
//    entonces el segundo botn anula este intento. NOTA: definir slo UN botn
//    por defecto en un formulario.
// Locate DefButton and cancel it behaviour
//  Limits: We cancel the first default button only. If there are more than one
//    defbutton (something wrong but sure that possible) then the seccond one
//    voids our try. NOTE: setup only ONE def pushbutton.
procedure TAutoCompleteEdit.OverrideDefault;
var
  Form: TCustomForm;
  idx: integer;

begin
  if DefButton <> nil then exit;

  Form := GetParentForm(Self);
  if Form <> nil then
    for idx := 0 to Form.ComponentCount - 1 do
      if Form.Components[idx] is TButton then begin
        DefButton := TButton(Form.Components[idx]);
        if DefButton.Default then begin
                               DefButton.Default := False;
                               break;
                             end else DefButton := nil;
      end;
end;

procedure TAutoCompleteEdit.ReleaseDefault;
begin
  if Assigned(DefButton) then DefButton.Default := True;
  DefButton := nil;
end;

procedure TAutoCompleteEdit.CalcSuggestPos(ControlRect: TRect);
var
  R: TRect;
  T, B: Integer;
  F: TCustomForm;
begin
  if bSpanParent then begin
    // Put it in the same coordinates space as the sugestions Listbox
    F := GetParentForm(Self);
    MapWindowPoints(Parent.Handle, F.Handle, ControlRect, 2);
  end;

  ListaSugerencias.Font := Font;
  ListaSugerencias.Color := backColor;
  SendMessage(ListaSugerencias.Handle, LB_GETITEMRECT, 0, Longint(@R));
  ListaSugerencias.Left := ControlRect.Left;
  ListaSugerencias.Width := ControlRect.Right - ControlRect.Left;
  ListaSugerencias.Height := uLines * ((R.Bottom - R.top)+ 2);

  // La lista de sugerencias tendr el mismo ancho que el control asociado
  // y estar posicionada justo debajo de el. Esto salvo que debajo no sea
  // visible por falta de espacio. En este caso se pone encima.
{$WARNINGS OFF}
  if bSpanParent then B := F.ClientHeight - StatusBarHeight(F)
                 else B := Parent.Height;
{$WARNINGS ON}

  B := B - ControlRect.Bottom - DELTA;
  T := ControlRect.Top - DELTA;
  if (B >= ListaSugerencias.Height) or (B >= T) then //  Cabe debajo ?
  begin
    // No exceder los lmites del padre
    // try not to span the parent edges
    T := uLines;
    While (T > 1) and (ListaSugerencias.Height > B) do begin
      Dec(T); ListaSugerencias.Height := T * ((R.Bottom - R.top)+ 2);
    end;
    ListaSugerencias.Top := ControlRect.Bottom + DELTA;
  end
  else if (T >= ListaSugerencias.Height) or (T > B) then begin //  Cabe Arriba ?
    // No exceder los lmites del padre
    // try not to span the parent edges
    B := uLines;
    While (B > 1) and (ListaSugerencias.Height > T) do begin
      Dec(B); ListaSugerencias.Height := B * ((R.Bottom - R.top)+ 2);
    end;
    ListaSugerencias.Top  := ControlRect.Top - ListaSugerencias.Height - DELTA;
  end;
end;

// Return the height of the given form's status bar if it has one.
function TAutoCompleteEdit.StatusBarHeight(Form: TCustomForm): integer;
var
  i: integer;
begin
  result := 0;
  for i := 0 to Form.ComponentCount - 1 do
    if Form.Components[i] is TStatusBar then begin
      result := TStatusBar(Form.Components[i]).Height;
      exit;
    end;
end;



{ TAutoCompleteDBEdit }

constructor TAutoCompleteDBEdit.Create(AOwner: TComponent);
begin
  inherited;

  uLines := 3;            // Suggestion box is 3 lines height by default
  bSuggest := True;       // By default do suggest
  bIgnoreChange := False;
  bFromList := False;
  backColor := RGB(255,255,180); // Kind of yellow

  iPos := -1;

  if (ListaSugerencias = nil) and not (csDesigning in ComponentState) then begin
    // Dejar que otro gestione la destruccin de la lista
    // Let other handle the disposal of the Listbox
    ListaSugerencias := TACListBox.Create(AOwner);
    ListaSugerencias.Parent := Self;
    ListaSugerencias.Name := 'AutoSuggestList';
    ListaSugerencias.Visible := False;
    ListaSugerencias.MultiSelect := False;
    ListaSugerencias.IntegralHeight := True;
    ListaSugerencias.TabStop := False;
    ListaSugerencias.Height := uLines * ListaSugerencias.ItemHeight;
  end;

  DictID := -1;
  if not (csDesigning in ComponentState) then begin
    if DictList = nil then DictList := TDictionaryList.Create;
    if FFichero <> '' then
      DictID := DictList.Add(ExtractFileName(FFichero), FFichero, True);
  end;
end;

destructor TAutoCompleteDBEdit.Destroy;
begin
    DictList.Delete(DictID);

  inherited;
end;

procedure TAutoCompleteDBEdit.SetColor(val: TColor);
begin
  if val <> backColor then begin
    backColor := val;
    if not (csDesigning in ComponentState) then
      ListaSugerencias.Color := val;
  end;
end;

function TAutoCompleteDBEdit.GetGroupIndex(c: char): integer;
var
 p1: PAnsiChar;
 L: TStringList;
begin
  L := DictList.GetDict(DictID);
  if c = '' then begin
    result := L.Count;
    exit;   // Avoid searching for nothing
  end;

  p1 := @c;
  for result := 0 to L.Count - 1 do begin
   try
     if CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
                         p1, 1, PChar(L[result]), 1) = 2 then
       exit;
   except
     break;
   end;
  end;
end;

procedure TAutoCompleteDBEdit.SetSuggestHeight(val: word);
begin
  if val <> uLines then begin
    uLines := val;
    if not (csDesigning in ComponentState) then
      ListaSugerencias.Height := uLines * ListaSugerencias.ItemHeight;
  end;
end;

procedure TAutoCompleteDBEdit.SetSuggest(val: boolean);
begin
  if val <> bSuggest then begin
    bSuggest := val;
    if not (csDesigning in ComponentState) then begin
      if val and not ListaSugerencias.Visible then Change;
      if not val then
        ListaSugerencias.Visible := False;
        if Assigned(FOnHideBox) then FOnHideBox(Self);
    end;
  end;
end;

procedure TAutoCompleteDBEdit.SetSpan(val: bool);
begin
  if val <> bSpanParent then begin
    bSpanParent := val;
  end;
end;

procedure TAutoCompleteDBEdit.SetDictionary(path: string);
begin
  if (FFichero <> path) then begin
    FFichero := path;
    if not (csDesigning in ComponentState) then
    begin
      DictList.Delete(DictID);
      DictID := DictList.Add('', path, True);
      if ListaSugerencias.Visible then Change;
    end;
  end;
end;

procedure TAutoCompleteDBEdit.WMWindowPosChanged(var Message: TWMWINDOWPOSCHANGED);
var
  rect: TRect;
begin
  inherited;

  if not (csDesigning in ComponentState) then begin
    rect.Top   := PWINDOWPOS(Message.WindowPos).y;
    rect.Bottom := rect.Top + PWINDOWPOS(Message.WindowPos).cy;
    rect.Left    := PWINDOWPOS(Message.WindowPos).x;
    rect.Right  := rect.top + PWINDOWPOS(Message.WindowPos).cx;
    CalcSuggestPos(rect);
  end;
end;

// El usuario ha pulsado una tecla especial. Nos interesan las flechas
// arriba y abajo para movernos por la lista de sugerencias y el retorno
// para seleccionar la sugerencia y hacerla activa.
//
// The user has pressed an special key. We are interested in up and down arrows
// to move through the suggestions list and return to select the actual one.
//
procedure TAutoCompleteDBEdit.CNKeyDown(var Message: TWMKeyDown);
begin
  if not (csDesigning in ComponentState) then begin
    if not ListaSugerencias.Visible then inherited
    else
    with Message do begin
      Result := 1;
      case CharCode of
        VK_UP, VK_DOWN, VK_RETURN, VK_ESCAPE: exit;
      end;
      Result := 0;
      inherited;
    end;
  end else inherited;
end;

procedure TAutoCompleteDBEdit.CNKeyUp(var Message: TWMKeyUp);
begin
  if not (csDesigning in ComponentState) then begin
    if not ListaSugerencias.Visible then inherited
    else
    with Message do begin
      Result := 0;
      case CharCode of
        VK_UP:
        begin
          if iPos = -1 then exit;
          if iPos > 0 then Dec(iPos)
                      else begin
                        iPos := -1;
                        ReleaseDefault;
                      end;
          ListaSugerencias.ItemIndex := iPos;
        end;

        VK_DOWN:
        begin
          if iPos = -1 then begin
                         iPos := 0;
                         OverrideDefault;
                       end else inc(iPos);
          if iPos >= ListaSugerencias.Items.Count then
            Dec(iPos);
          ListaSugerencias.ItemIndex := iPos;
        end;

        VK_RETURN:
        if ((iPos <> -1) and (iPos < ListaSugerencias.Items.Count)) then
        begin
          // Esto va a disparar un mensaje de cambio. Hacer que no nos moleste
          // This is firing a Change message, so don't let it disturb us...
          bIgnoreChange := True;
          Text := ListaSugerencias.Items[iPos];
          bFromList := True;
          bIgnoreChange := False;
          ListaSugerencias.Visible := False;
          if Assigned(FOnHideBox) then FOnHideBox(Self);
          SelStart := Length(Text);
          ReleaseDefault;
        end;

        VK_ESCAPE:
        begin
          ListaSugerencias.Visible := False;
          if Assigned(FOnHideBox) then FOnHideBox(Self);
          ReleaseDefault;
        end;
      end;
    end;
  end else inherited;
end;

// El contenido del control ha cambiado. Actualizar la lista de sugerencias
//
// The control's contents has changed. Update suggestions list
procedure TAutoCompleteDBEdit.Change;
var
  idx: integer;
  L: word;
  Lista: TStringList;

begin
  if not (csDesigning in ComponentState) and (GetFocus = Handle) and
     (DictID <> -1) and bSuggest and not bIgnoreChange then begin
    ReleaseDefault; iPos := -1;
    ListaSugerencias.Clear;
    L := Length(Text);
    if L > 0 then begin
      // Seleccionar el grupo de la letra inicial.
      // Select starting letter group
      if (L =1) or (iGrupo = 0) then iGrupo := GetGroupIndex(Text[1]);
      Lista := DictList.GetDict(DictID);
      if Lista <> nil then
        for idx := iGrupo to Lista.Count - 1 do
        begin
          if CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
                           PChar(Lista[idx]), 1, PChar(Text), 1) = 3 then
            break; // Nos hemos pasado de letra
          if L > Length(Lista[idx]) then continue; // Una subcadena
          if CompareString(LOCALE_USER_DEFAULT, NORM_IGNORECASE,
                           PChar(Text), L, PChar(Lista[idx]), L) = 2 then
            ListaSugerencias.Items.Add(Lista[idx]);
        end;

      if (ListaSugerencias.Items.Count > 0) then begin
        ListaSugerencias.Visible := True;
        ListaSugerencias.BringToFront;
        if Assigned(FOnShowBox) then FOnShowBox(Self, ListaSugerencias.Items.Count);
      end else begin
        ListaSugerencias.Visible := False;
        if Assigned(FOnHideBox) then FOnHideBox(Self);
      end;
    end else begin
      ListaSugerencias.Visible := False;
      if Assigned(FOnHideBox) then FOnHideBox(Self);
    end;
  end;

  inherited;
end;

procedure TAutoCompleteDBEdit.CMEnter(var Message: TCMEnter);
begin
  if not (csDesigning in ComponentState) then begin
    // Asociarla al contexto correspondiente
    // Link to corresponding context
    if bSpanParent then ListaSugerencias.Parent := GetParentForm(Self)
                   else ListaSugerencias.Parent := Parent;

    // Ajustarla al control de que se trate
    // Adjust to control's size
    CalcSuggestPos(BoundsRect);
  end;

  inherited;
end;

procedure TAutoCompleteDBEdit.DoExit;
begin
  if not (csDesigning in ComponentState) then begin
    // Salvar el valor actual del control en la lista asociada
    // si no ha sido un valor obtenido de ella.
    if bSuggest and not bFromList and (Length(Text) > 0) then
      DictList.AddWord(DictID, Text);
    bFromList := False;

    if ListaSugerencias.Visible then begin
      ListaSugerencias.Visible := False;
      if Assigned(FOnHideBox) then FOnHideBox(Self);
    end;
  end;

  inherited;
end;

// Localizar el botn por defecto y cancelar su comportamiento.
//  Limitaciones: Slo cancela el primer botn por defecto localizado . Si hay
//    ms de una en un formulario (algo errneo pero parece ser que posible)
//    entonces el segundo botn anula este intento. NOTA: definir slo UN botn
//    por defecto en un formulario.
//
// Locate DefButton and cancel its behaviour
//  Constraints: We cancel the first default button only. If there are more than
//    one defbutton (wrong but pretty possible) then the second one voids our
//    try. FIX: setup only ONE def pushbutton.
//
procedure TAutoCompleteDBEdit.OverrideDefault;
var
  Form: TCustomForm;
  idx: integer;

begin
  if DefButton <> nil then exit;

  Form := GetParentForm(Self);
  if Form <> nil then
    for idx := 0 to Form.ComponentCount - 1 do
      if Form.Components[idx] is TButton then begin
        DefButton := TButton(Form.Components[idx]);
        if DefButton.Default then begin
                               DefButton.Default := False;
                               break;
                             end else DefButton := nil;
      end;
end;

procedure TAutoCompleteDBEdit.ReleaseDefault;
begin
  if Assigned(DefButton) then DefButton.Default := True;
  DefButton := nil;
end;

procedure TAutoCompleteDBEdit.CalcSuggestPos(ControlRect: TRect);
var
  R: TRect;
  T, B: Integer;
  F: TCustomForm;
begin
  if bSpanParent then begin
    // Make it relative to same coordinates space asthe suggestions listbox
    F := GetParentForm(Self);
    MapWindowPoints(Parent.Handle, F.Handle, ControlRect, 2);
  end;

  ListaSugerencias.Font := Font; // Level visual appearance
  ListaSugerencias.Color := backColor;
  SendMessage(ListaSugerencias.Handle, LB_GETITEMRECT, 0, Longint(@R));
  ListaSugerencias.Left := ControlRect.Left;
  ListaSugerencias.Width := ControlRect.Right - ControlRect.Left;
  ListaSugerencias.Height := uLines * ((R.Bottom - R.top)+ 2);

  // La lista de sugerencias tendr el mismo ancho que el control asociado
  // y estar posicionada justo debajo de el. Esto salvo que debajo no sea
  // visible por falta de espacio. En este caso se pone encima.
{$WARNINGS OFF}
  if bSpanParent then B := F.ClientHeight - StatusBarHeight(F)
                 else B := Parent.Height;
{$WARNINGS ON}

  B := B - ControlRect.Bottom - DELTA;
  T := ControlRect.Top - DELTA;
  if (B >= ListaSugerencias.Height) or (B >= T) then //  Cabe debajo ?
  begin
    // No exceder los lmites del padre
    // try not to span the parent edges
    T := uLines;
    While (T > 1) and (ListaSugerencias.Height > B) do begin
      Dec(T); ListaSugerencias.Height := T * ((R.Bottom - R.top)+ 2);
    end;
    ListaSugerencias.Top  := ControlRect.Bottom + DELTA;
  end
  else if (T >= ListaSugerencias.Height) or (T > B) then begin //  Cabe Arriba ?
    // No exceder los lmites del padre
    // try not to span the parent edges
    B := uLines;
    While (B > 1) and (ListaSugerencias.Height > T) do begin
      Dec(B); ListaSugerencias.Height := B * ((R.Bottom - R.top)+ 2);
    end;
    ListaSugerencias.Top  := ControlRect.Top - ListaSugerencias.Height - DELTA;
  end;
end;

// Return the height of the given form's status bar if it has one.
function TAutoCompleteDBEdit.StatusBarHeight(Form: TCustomForm): integer;
var
  i: integer;
begin
  result := 0;
  for i := 0 to Form.ComponentCount - 1 do
    if Form.Components[i] is TStatusBar then begin
      result := TStatusBar(Form.Components[i]).Height;
      exit;
    end;
end;

procedure Register;
begin
  RegisterComponents('Additional', [TAutoCompleteEdit]);
  RegisterComponents('Data Controls', [TAutoCompleteDBEdit]);
end;

initialization

finalization
  if Assigned(DictList) then DictList.Free;
end.



